﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=7.8
@EndOfDesignText@
Sub Class_Globals
	Public BANano As BANano 'ignore
	Private mCachedPages As Map
	
	Private mPort As Int = 55055 '80
	Private mPortSSL As Int = 0 '55443 '53001
	
	Private mAppVersion As String = DateTime.now
	
	Private mCacheScavengePeriodSeconds As Int = 15*60 ' 15 minutes
	Private mSessionMaxInactiveIntervalSeconds As Int = 15*60 ' -1 = immortal but beware! This means the cache is NEVER emptied!
	
	Private mOutputFolder As String = File.DirApp
	Private mAppName As String = "BANanoServer"
	
	Private mRootFilter As String = "SERVERRootFilter"
	Private mErrorHandler As String = "SERVERErrorHandler"
	Private mHttp2HttpsFilter As String = "SERVERHttp2HttpsFilter"		
	Private mUploadHandler As String = "SERVERUploadHandler"
	Private mUploadHandlerPath As String = "/upload"
	Private mBROWSERPrefix As String = "BROWSER"
	Private mSHAREDPrefix As String = "SHARED"
	
	Private mStartPage As String
	Private mUploadFolder As String = File.DirApp
	Private mUploadMaxSize As Long = -1
	Private mUploadAllowedFileTypes As String = ""
	
	Private mSrvr As Server
	Private mSockets As List
	Private mHandlers As List
	Private mFilters As List
	Private mBackgroundWorkers As List
	Private mDoSFilters As List
	Private mDoSFiltersOptions As List
	Private mAuthFilters As List	
	
	Private mLoadBalancerHost As String
	Private mLoadBalancerPort As Int = 80
	Private mHost As String = "localhost"
	
	Public WSPaths As Map
End Sub

'Initializes the server.  Should be done in Main.
Public Sub Initialize(INIFileFullPath As String)
	WSPaths.Initialize
	If INIFileFullPath <> "" Then
		If File.Exists("", INIFileFullPath) Then
			Dim INIInfo As List = File.ReadList("", INIFileFullPath)
			For i = 0 To INIInfo.Size - 1
				Dim line As String = INIInfo.Get(i)
				If line.StartsWith("Host=") Then
					mHost = line.SubString("Host=".Length)
				End If
				If line.StartsWith("Port=") Then
					mPort = line.SubString("Port=".Length)
				End If
				If line.StartsWith("PortSSL=") Then
					mPortSSL = line.SubString("PortSSL=".Length)
				End If
				If line.StartsWith("LoadBalancerHost=") Then
					mLoadBalancerHost = line.SubString("LoadBalancerHost=".Length)
				End If
				If line.StartsWith("LoadBalancerPort=") Then
					mLoadBalancerPort = line.SubString("LoadBalancerPort=".Length)
				End If
				If line.StartsWith("CacheScavengePeriodSeconds=") Then
					mCacheScavengePeriodSeconds = line.SubString("CacheScavengePeriodSeconds=".Length)
				End If
				If line.StartsWith("SessionMaxInactiveIntervalSeconds=") Then
					mSessionMaxInactiveIntervalSeconds = line.SubString("SessionMaxInactiveIntervalSeconds=".Length)
				End If				
			Next
		End If	
	End If	
	' check if the loadbalancer is running
	If mLoadBalancerHost <> "" Then
		Dim Host As String = mLoadBalancerHost.Replace("http://", "").Replace("https://", "")
		If BANano.PingServer(Host, mLoadBalancerPort, 10000) = False Then
			mLoadBalancerHost = ""
		End If
	End If
	BANano.IsForBANanoServer = True
	BANano.UploadHandlerPath = mUploadHandlerPath
	BANano.TranspilerOptions.SetSessionMaxInactiveInterval(mSessionMaxInactiveIntervalSeconds)
	
	BANano.BROWSERPrefix = mBROWSERPrefix
	BANano.SHAREDPrefix = mSHAREDPrefix
End Sub

' opens an url, windows only! 
Public Sub OpenURL(url As String)
	Dim NativeMe As JavaObject
	NativeMe = Me
	NativeMe.RunMethod("RunUrl", Array("rundll32 url.dll,FileProtocolHandler " & Chr(34) & url & Chr(34)))
End Sub

#if JAVA
	import java.io.IOException;
	
	public void RunUrl(String url) {
		String os = System.getProperty("os.name").toLowerCase();
		if (os.indexOf("win") >= 0) {
			try {
				Runtime rt = Runtime.getRuntime();
				rt.exec(url);
			} catch(IOException ioe) {
				ioe.printStackTrace();
			}
		}
	}
#End If

' Register this server with the BANanoLoadBalancer app that it is available for requests
public Sub RegisterMeToLoadBalancer()
	If mLoadBalancerHost <> "" Then
		Dim json As JSONGenerator
		Dim m As Map
		m.Initialize
		m.Put("host", mHost)
		If mPortSSL = 0 Then
			m.Put("protocol", "http://")
			m.Put("port", mPort)
		Else
			m.Put("protocol", "https://")
			m.Put("port", mPortSSL)
		End If			
		json.Initialize(m)
		
		Dim j As HttpJob
		j.Initialize("", Me)
		j.PostString(Main.Server.LoadBalancerHost & ":" & Main.Server.LoadBalancerPort & "/BANRegister", json.ToString)
		Try
			j.GetRequest.SetContentType("application/json")
			Wait For (j) JobDone(j As HttpJob)
			If j.Success Then
				'Log("Success")
			Else
				Log(j.ErrorMessage)
			End If
		Catch
			Log(LastException.Message)
		End Try
		j.Release
	End If
End Sub

' return s the base URL (protocol + host + port)
' e.g. https://localhost:55055
public Sub GetBaseURL() As String
	If mPortSSL > 0 Then
		Return "https://" & mHost & ":" & mPortSSL
	Else
		Return "http://" & mHost & ":" & mPort
	End If
End Sub

' Gets/Sets the host name.  Default is localhost
public Sub getHost() As String
	Return mHost
End Sub

public Sub setHost(host As String)
	mHost = host
End Sub

' Gets/Sets the URL to the BANanoLoadBalancer
' e.g. http://localhost
public Sub getLoadBalancerHost() As String
	Return mLoadBalancerHost
End Sub

public Sub setLoadBalancerHost(loadBalancerHost As String)
	mLoadBalancerHost = loadBalancerHost
	If mLoadBalancerHost <> "" Then
		Dim Host As String = mLoadBalancerHost.Replace("http://", "").Replace("https://", "")
		If BANano.PingServer(Host, mLoadBalancerPort, 10000) = False Then
			mLoadBalancerHost = ""
		End If
	End If
End Sub

' Gets/Sets the port on which the LoadBalancer is running
public Sub getLoadBalancerPort() As Int
	Return mLoadBalancerPort
End Sub

public Sub setLoadBalancerPort(loadBalancerPort As Int)
	mLoadBalancerPort = loadBalancerPort	
End Sub

' Gets the map with all the Cached Pages
Public Sub getCachedPages As Map
	Return mCachedPages
End Sub

' Gets/Sets the http port. Default is 55055
public Sub getPort() As Int
	Return mPort
End Sub

public Sub setPort(port As Int)
	mPort = port
End Sub

' Gets/Sets the https port.  Default is 0 (not set)
public Sub getPortSSL()  As Int
	Return mPortSSL
End Sub

public Sub setPortSSL(portSSL As Int)
	mPortSSL = portSSL
End Sub

' Gets/Sets the app version. Default is DateTime.Now
public Sub getAppVersion() As String
	Return mAppVersion
End Sub

public Sub setAppVersion(appVersion As String)
	mAppVersion = appVersion
End Sub

' Gets/Sets the app name. Default is BANanoServer
public Sub getAppName() As String
	Return mAppName
End Sub

public Sub setAppName(appName As String)
	mAppName = appName
End Sub

' Gets/Sets the time in seconds the Scavenger must run.  Will remove disconnected pages from CachedPages.  Default is 900 (15 minutes)
public Sub getCacheScavengePeriodSeconds() As Int
	Return mCacheScavengePeriodSeconds
End Sub

public Sub setCacheScavengePeriodSeconds(cacheScavengePeriodSeconds As Int)
	mCacheScavengePeriodSeconds = cacheScavengePeriodSeconds
End Sub

' Get/Sets the time in seconds a Session is considered active. Default is 900 (15 minutes)
public Sub getSessionMaxInactiveIntervalSeconds() As Int
	Return mSessionMaxInactiveIntervalSeconds
End Sub

public Sub setSessionMaxInactiveIntervalSeconds(sessionMaxInactiveIntervalSeconds As Int)
	mSessionMaxInactiveIntervalSeconds = sessionMaxInactiveIntervalSeconds
	BANano.TranspilerOptions.SetSessionMaxInactiveInterval(mSessionMaxInactiveIntervalSeconds)
End Sub

' Get/Sets where BANano has to generate the StaticFolder.  Is als the same place where the BANanoServer gets its files.
' default is File.DirApp
public Sub getOutputFolder() As String
	Return mOutputFolder
End Sub

public Sub setOutputFolder(outputFolder As String)
	mOutputFolder = outputFolder
End Sub

' Gets/Sets the root folder filter class (path / ) Default is SERVERRootFilter
public Sub getRootFilter() As String
	Return mRootFilter
End Sub

public Sub setRootFilter(rootFilter As String)
	mRootFilter = rootFilter
End Sub

' Gets/Sets the error handler class (path /YourStaticFolder/error ) Default is SERVERErrorHandler
public Sub getErrorHandler() As String
	Return mErrorHandler
End Sub

public Sub setErrorHandler(errorHandler As String)
	mErrorHandler = errorHandler
End Sub

' Gets/Sets the http to https (http2) filter class (path /* ) Default is SERVERHttp2HttpsFilter
public Sub getHttp2HttpsFilter() As String
	Return mHttp2HttpsFilter
End Sub

public Sub setHttp2HttpsFilter(http2HttpsFilter As String)
	mHttp2HttpsFilter = http2HttpsFilter
End Sub

' Get/Sets the prefix of your BANano BROWSER classes.  Default is BROWSER
public Sub getBROWSERPrefix() As String
	Return mBROWSERPrefix
End Sub

public Sub setBROWSERPrefix(BROWSERPrefix As String)
	mBROWSERPrefix = BROWSERPrefix	
	BANano.BROWSERPrefix = mBROWSERPrefix	
End Sub

' Get/Sets the prefix of your BANano SHARED classes.  Default is SHARED
public Sub getSHAREDPrefix() As String
	Return mSHAREDPrefix
End Sub

public Sub setSHAREDPrefix(SHAREDPrefix As String)
	mSHAREDPrefix = SHAREDPrefix
	BANano.SHAREDPrefix = mSHAREDPrefix
End Sub

' Get/Sets the Start page the root filter should go to.  Default is empty
public Sub getStartPage() As String
	Return mStartPage
End Sub

public Sub setStartPage(startPage As String)
	mStartPage = startPage
End Sub

' Get/Sets the upload folder
public Sub getUploadFolder() As String
	Return mUploadFolder
End Sub

public Sub setUploadFolder(uploadFolder As String)
	mUploadFolder = uploadFolder
End Sub

' Gets/Sets the maximum allowed file size. Default is -1, unlimited
public Sub getUploadMaxSize() As Long
	Return mUploadMaxSize
End Sub

public Sub setUploadMaxSize(uploadMaxSize As Long)
	mUploadMaxSize = uploadMaxSize
End Sub

' Gets/Sets the allowed file types. Default is "", all
' is a comma delimited extension list
' e.g. PNG;JPG;JPEG
public Sub getUploadAllowedFileTypes() As String
	Return mUploadAllowedFileTypes
End Sub

public Sub setUploadAllowedFileTypes(uploadAllowedFileTypes As String)
	mUploadAllowedFileTypes = uploadAllowedFileTypes
End Sub

' Gets/Sets the upload handler.  Default is SERVERUploadHandler
public Sub getUploadHandler() As String
	Return mUploadHandler
End Sub

public Sub setUploadHandler(uploadHandler As String)
	mUploadHandler = uploadHandler
End Sub

' Gets/Sets the upload handler path.  Default is /upload
public Sub getUploadHandlerPath() As String
	Return mUploadHandlerPath
End Sub

public Sub setUploadHandlerPath(uploadHandlerPath As String)
	BANano.UploadHandlerPath = mUploadHandlerPath
	mUploadHandlerPath = uploadHandlerPath
End Sub

' Maps a WebSocket class.
' Path - The WebSocket will be mapped To this path.
' Class - WebSocket class (string).
public Sub AddWebSocket(Path As String, Class As String)
	If mSockets.IsInitialized = False Then
		mSockets.Initialize
	End If
	mSockets.add(Path & ";" & Class)
	Dim Page As String
	Page = Path
	If Page.EndsWith("/") Then
		Page = Page.SubString2(0, Page.Length-1)
	End If
	Dim i As Int = Page.LastIndexOf("/")
	If i > 0 Then
		Page = Page.SubString(i+1)
	End If
	Dim MappedHTML As String = mBROWSERPrefix.ToLowerCase & Class.SubString(mBROWSERPrefix.Length-1).ToLowerCase
	WSPaths.Put(mBROWSERPrefix & Page, MappedHTML)
	If WSPaths.Size = 1 And mStartPage = "" Then
		mStartPage = Page.ToLowerCase
	End If
End Sub

' Maps a handler class.
' Path - The handler will be mapped To the following Path. You can use wildcards in the Path.
' Class - The Class name (string).
' SingleThreadHandler - Whether this handler should always run in the Main thread.
public Sub AddHandler(Path As String, Class As String, SingleThreadHandler As Boolean)
	If mHandlers.IsInitialized = False Then
		mHandlers.Initialize
	End If
	mHandlers.add(Path & ";" & Class & ";" & SingleThreadHandler)
End Sub

' Maps a filter class.
' Path - The filter will be mapped To the following Path. You can use wildcards in the Path.
' Class - The Class name (string).
 'SingleThreadHandler - Whether this handler should always run in the Main thread.
public Sub AddFilter(Path As String, Class As String, SingleThreadHandler As Boolean)
	If mFilters.IsInitialized = False Then
		mFilters.Initialize
	End If
	mFilters.add(Path & ";" & Class & ";" & SingleThreadHandler)
End Sub

' Adds a background worker. An instance of the specified class will be created and initialized from a background thread.
' Class - The Class name. The Class should be a standard Class (Not server handler).
'
' You can call StartMessageLoop in the Initialize sub to keep the class instance running for as long as needed.
' Note that in debug mode the code will be executed by the Main thread.
public Sub AddBackgroundWorker(Class As String)
	If mBackgroundWorkers.IsInitialized = False Then
		mBackgroundWorkers.Initialize
	End If
	mBackgroundWorkers.add(Class)
End Sub

' Adds a filter that protects against intentional or unintentional denial of service attacks.
' It restricts the number of requests sent from the same connection.
' Path - The filter will be mapped To the given path. You can use wildcards in the path.
' Settings - An optional map with various settings. See this link: http://www.eclipse.org/jetty/documentation/9.4.x/dos-filter.html
'
' Example:srvr.AddDoSFilter("/*", Null)
public Sub AddDoSFilter(Path As String, Settings As Map)
	If mDoSFilters.IsInitialized = False Then
		mDoSFilters.Initialize
		mDoSFiltersOptions.Initialize
	End If
	mDoSFilters.add(Path)
	mDoSFiltersOptions.add(Settings)
End Sub

' Add an Auth filter class
' This is an example of such an Auth filter, using the ABJJWT library to make and check Tokens
'
' <code>
' 'Filter class
' Sub Class_Globals
' End Sub
' 
' Public Sub Initialize
' 
' End Sub
' 
' 'Return True to allow the request to proceed.
' Public Sub Filter(req As ServletRequest, resp As ServletResponse) As Boolean
' 	Dim Token As String = req.GetHeader("api_key")
' 	resp.ContentType = "application/json"
' 	resp.SetHeader("X-Frame-Options", "DENY")
' 	resp.SetHeader("X-XSS-Protection", "1;mode=block")
' 	resp.SetHeader("Strict-Transport-Security", "max-age=31536000;includeSubDomains;preload")
' 	resp.SetHeader("X-Content-Type-Options", "nosniff")
' 	resp.SetHeader("Referrer-Policy", "no-referrer-when-downgrade")
' 	resp.SetHeader("Content-Security-Policy", "script-src https://yourdomain.com")
' 	resp.SetHeader("Feature-Policy", "microphone 'none'")
' 	If Token <> "" Then
' 		Dim jjwt As ABJJWT
' 		Dim TokenJJWT As ABJJWTToken = jjwt.CheckToken(Token, "YOURMAGICKEY",0)
' 		If TokenJJWT = Null Then
' 			resp.Status = 401
' 			resp.Write("Unauthorized")
' 			Return False
' 		End If		
' 	Else
' 		resp.Status = 401
' 		resp.Write("Unauthorized")
' 		Return False		
' 	End If
' 	resp.SetHeader("Access-Control-Allow-Origin","*")
' 	resp.SetHeader("Access-Control-Allow-Methods" ,"GET, POST, UPDATE, DELETE, OPTIONS")
' 	resp.SetHeader("Access-Control-Allow-Headers", "Access-Control-Allow-Headers, Origin, Accept, X-Requested-With, Content-Type, Access-Control-Request-Method, Access-Control-Request-Headers, Authorization")
' 	Return True
' End Sub
' </code>
public Sub AddAuthFilter(Path As String, Class As String, SingleThreadHandler As Boolean)
	If mAuthFilters.IsInitialized = False Then
		mAuthFilters.Initialize
	End If
	mAuthFilters.add(Path & ";" & Class & ";" & SingleThreadHandler)
End Sub

' Starts the server. The handlers and the configuration settings should be set before the server is started.
public Sub StartServer()
	' start the server
	mSrvr.Initialize("srvr")
	
	If mUploadFolder = File.DirApp Then
		File.MakeDir(mUploadFolder,BANano.StaticFolder & "/uploads")
		mUploadFolder = mUploadFolder & "/" & BANano.StaticFolder & "/uploads"
	End If
	
	If mRootFilter <> "" Then
		mSrvr.AddFilter("/", mRootFilter, False )
	End If
	mSrvr.AddFilter("/*", "SERVERCacheControl", False)
	
	If mErrorHandler <> "" Then
		mSrvr.SetCustomErrorPages(CreateMap("org.eclipse.jetty.server.error_page.global": "/error")) ' OPTIONAL
		mSrvr.AddHandler("/error", "SERVERErrorHandler", False) ' OPTIONAL
	End If
	
	' the global page filter which will build de BANano page en connect with the SERVER page through a websocket
	mSrvr.AddFilter("/*", "SERVERPageFilter", False)
	
	If mDoSFilters.IsInitialized Then
		For i = 0 To mDoSFilters.Size - 1
			Dim s As String = mDoSFilters.Get(i)
			Dim Settings As Map = mDoSFiltersOptions.get(i)
			mSrvr.AddDoSFilter(s, Settings)
		Next
	End If
	
	If mAuthFilters.IsInitialized Then
		For Each s As String In mAuthFilters
			Dim spl() As String = Regex.Split(";", s)
			If spl(2).ToLowerCase = "true" Then
				mSrvr.AddFilter(spl(0), spl(1), True)
			Else
				mSrvr.AddFilter(spl(0), spl(1), False)
			End If
		Next
	End If
	
	If mFilters.IsInitialized Then
		For Each s As String In mFilters
			Dim spl() As String = Regex.Split(";", s)
			If spl(2).ToLowerCase = "true" Then
				mSrvr.AddFilter(spl(0), spl(1), True)
			Else
				mSrvr.AddFilter(spl(0), spl(1), False)
			End If
		Next
	End If
	
	mSrvr.AddHandler("/upload", "SERVERUploadHandler", False)
	
	If mHandlers.IsInitialized Then
		For Each s As String In mHandlers
			Dim spl() As String = Regex.Split(";", s)
			If spl(2).ToLowerCase = "true" Then
				mSrvr.AddHandler(spl(0), spl(1), True)
			Else
				mSrvr.AddHandler(spl(0), spl(1), False)
			End If
		Next
	End If
	
	mSrvr.AddHandler("/BANStatus", "SERVERLoadBalancerHandler", False)
		
	' your own pages, the SERVER classes from the type Websocket
	'Srvr.AddWebSocket("/ws/" & BANano.GetStaticFolder & "/about" , "SERVERAbout")
	If mSockets.IsInitialized Then
		For Each s As String In mSockets
			Dim spl() As String = Regex.Split(";", s)
			mSrvr.AddWebSocket(spl(0) , spl(1))
		Next
	End If
	
	mSrvr.StaticFilesFolder = File.Combine(mOutputFolder, "/" & BANano.StaticFolder)
	mSrvr.Port = mPort
	mSrvr.LogWaitingMessages = False
		
	#If RELEASE		
		mSrvr.SetStaticFilesOptions(CreateMap("gzip":True,"dirAllowed":False))
		mSrvr.GzipEnabled = True
	#Else		
		mSrvr.SetStaticFilesOptions(CreateMap("gzip":False,"dirAllowed":False))
	#End If
	
	mSrvr.AddBackgroundWorker("SERVERCacheScavenger")
	
	If mBackgroundWorkers.IsInitialized Then
		For Each s As String In mBackgroundWorkers
			mSrvr.AddBackgroundWorker(s)
		Next
	End If
	
	mSrvr.Start
	
	' setting the app cookie
	Dim joServer As JavaObject = mSrvr
	joServer.GetFieldJO("server").RunMethod("stop", Null)
	joServer.GetFieldJO("context").RunMethodJO("getSessionHandler", Null).RunMethodJO("getSessionCookieConfig", Null).RunMethod("setMaxAge", Array(31536000)) ' 1 year
	joServer.GetFieldJO("context").RunMethodJO("getSessionHandler", Null).RunMethodJO("getSessionCookieConfig", Null).RunMethod("setName", Array(mAppName.ToUpperCase))
	joServer.GetFieldJO("server").RunMethod("start", Null)
	
	' setting de time the scavenger should cleanup sessions
	Dim secs As Long = mCacheScavengePeriodSeconds ' must be defined as a long, else you get a 'java.lang.RuntimeException: Method: setIntervalSec not matched.' error
	joServer.GetFieldJO("context").RunMethodJO("getSessionHandler", Null).RunMethodJO("getSessionIdManager", Null).RunMethodJO("getSessionHouseKeeper", Null).RunMethod("setIntervalSec", Array As Object(secs))
	
	' setting the timeout for a connection
	Dim jo As JavaObject = mSrvr
	Dim connectors() As Object = jo.GetFieldJO("server").RunMethod("getConnectors", Null)
	Dim timeout As Long = mSessionMaxInactiveIntervalSeconds*1000
	For Each c As JavaObject In connectors
		c.RunMethod("setIdleTimeout", Array(timeout))
	Next
	
	mCachedPages = mSrvr.CreateThreadSafeMap
End Sub

' Starts the server in https/http2 mode. The handlers and the configuration settings should be set before the server is started.
public Sub StartServerHTTP2(SSLKeyStoreFileName As String, SSLKeyStorePassword As String, SSLKeyManagerPassword As String)
	Dim ssl As SslConfiguration
	ssl.Initialize
	ssl.SetKeyStorePath(File.DirApp, SSLKeyStoreFileName) 'path to keystore file
	ssl.KeyStorePassword = SSLKeyStorePassword
	ssl.KeyManagerPassword = SSLKeyManagerPassword
	mSrvr.SetSslConfiguration(ssl, mPortSSL)

	' start the server
	mSrvr.Initialize("srvr")
	
	If mUploadFolder = File.DirApp Then
		File.MakeDir(mUploadFolder,BANano.StaticFolder & "/uploads")
		mUploadFolder = mUploadFolder & "/" & BANano.StaticFolder & "/uploads"
	End If
	
	If mRootFilter <> "" Then
		mSrvr.AddFilter("/", mRootFilter, False )
	End If
	If mHttp2HttpsFilter <> "" Then
		mSrvr.AddFilter("/*", mHttp2HttpsFilter <> "", False )
	End If
	mSrvr.AddFilter("/*", "SERVERCacheControl", False)
	
	If mErrorHandler <> "" Then
		mSrvr.SetCustomErrorPages(CreateMap("org.eclipse.jetty.server.error_page.global": "/error")) ' OPTIONAL
		mSrvr.AddHandler("/error", "SERVERErrorHandler", False) ' OPTIONAL
	End If
	
	' the global page filter which will build de BANano page en connect with the SERVER page through a websocket
	mSrvr.AddFilter("/*", "SERVERPageFilter", False)
	
	If mDoSFilters.IsInitialized Then
		For i = 0 To mDoSFilters.Size - 1
			Dim s As String = mDoSFilters.Get(i)
			Dim Settings As Map = mDoSFiltersOptions.get(i)
			mSrvr.AddDoSFilter(s, Settings)
		Next
	End If
	
	If mAuthFilters.IsInitialized Then
		For Each s As String In mAuthFilters
			Dim spl() As String = Regex.Split(";", s)
			If spl(2).ToLowerCase = "true" Then
				mSrvr.AddFilter(spl(0), spl(1), True)
			Else
				mSrvr.AddFilter(spl(0), spl(1), False)
			End If
		Next
	End If
	
	' your own pages, the SERVER classes from the type Websocket
	'Srvr.AddWebSocket("/ws/" & BANano.GetStaticFolder & "/about" , "SERVERAbout")	
	If mFilters.IsInitialized Then
		For Each s As String In mFilters
			Dim spl() As String = Regex.Split(";", s)
			If spl(2).ToLowerCase = "true" Then
				mSrvr.AddFilter(spl(0), spl(1), True)
			Else
				mSrvr.AddFilter(spl(0), spl(1), False)
			End If
		Next
	End If
	
	mSrvr.AddHandler("/upload", "SERVERUploadHandler", False)
	
	If mHandlers.IsInitialized Then
		For Each s As String In mHandlers
			Dim spl() As String = Regex.Split(";", s)
			If spl(2).ToLowerCase = "true" Then
				mSrvr.AddHandler(spl(0), spl(1), True)
			Else
				mSrvr.AddHandler(spl(0), spl(1), False)
			End If
		Next
	End If
	
	mSrvr.AddHandler("/BANStatus", "SERVERLoadBalancerHandler", False)
		
	' your own pages, the SERVER classes from the type Websocket
	'mSrvr.AddWebSocket("/ws/" & BANano.GetStaticFolder & "/about" , "SERVERAbout")
	If mSockets.IsInitialized Then
		For Each s As String In mSockets
			Dim spl() As String = Regex.Split(";", s)
			mSrvr.AddWebSocket(spl(0) , spl(1))
		Next
	End If
	
	mSrvr.StaticFilesFolder = File.Combine(mOutputFolder, "/" & BANano.StaticFolder)
	mSrvr.Port = mPort
	mSrvr.LogWaitingMessages = False
	mSrvr.Http2Enabled = True
	#If RELEASE		
		mSrvr.SetStaticFilesOptions(CreateMap("gzip":True,"dirAllowed":False))
		mSrvr.GzipEnabled = True
	#Else		
	mSrvr.SetStaticFilesOptions(CreateMap("gzip":False,"dirAllowed":False))
	#End If
	
	mSrvr.AddBackgroundWorker("SERVERCacheScavenger")
	
	If mBackgroundWorkers.IsInitialized Then
		For Each s As String In mBackgroundWorkers
			mSrvr.AddBackgroundWorker(s)
		Next
	End If
	
	mSrvr.Start
	
	' setting the app cookie
	Dim joServer As JavaObject = mSrvr
	joServer.GetFieldJO("server").RunMethod("stop", Null)
	joServer.GetFieldJO("context").RunMethodJO("getSessionHandler", Null).RunMethodJO("getSessionCookieConfig", Null).RunMethod("setMaxAge", Array(31536000)) ' 1 year
	joServer.GetFieldJO("context").RunMethodJO("getSessionHandler", Null).RunMethodJO("getSessionCookieConfig", Null).RunMethod("setName", Array(mAppName.ToUpperCase))
	joServer.GetFieldJO("server").RunMethod("start", Null)
	
	' setting de time the scavenger should cleanup sessions
	Dim secs As Long = mCacheScavengePeriodSeconds ' must be defined as a long, else you get a 'java.lang.RuntimeException: Method: setIntervalSec not matched.' error
	joServer.GetFieldJO("context").RunMethodJO("getSessionHandler", Null).RunMethodJO("getSessionIdManager", Null).RunMethodJO("getSessionHouseKeeper", Null).RunMethod("setIntervalSec", Array As Object(secs))
	
	' setting the timeout for a connection
	Dim jo As JavaObject = mSrvr
	Dim connectors() As Object = jo.GetFieldJO("server").RunMethod("getConnectors", Null)
	Dim timeout As Long = mSessionMaxInactiveIntervalSeconds*1000
	For Each c As JavaObject In connectors
		c.RunMethod("setIdleTimeout", Array(timeout))
	Next
	
	mCachedPages = mSrvr.CreateThreadSafeMap
End Sub

' Navigates to a certain url in the browser.
'
' ws - a server websocket
' PageID - the BANPageID from thecalling page (to remove it from the CachedPages)
' TargetUrl - the url to navigate to
' OpenInNewTab - open the page in a new tab in the browser
Public Sub NavigateToPage(ws As WebSocket, PageId As String, TargetUrl As String, OpenInNewTab As Boolean)
	If PageId.Length > 0 Then BANano.B4JRemoveMeFromCache(mCachedPages, PageId)
	If ws.Open Then
		If OpenInNewTab Then
			Dim s As String
			' check if a mobile phone or a tablet
			s = $"var check = false;
  			(function(a){if(/(android|bb\d+|meego).+mobile|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows ce|xda|xiino|android|ipad|playbook|silk/i.test(a)||/1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i.test(a.substr(0,4))) check = true;})(navigator.userAgent||navigator.vendor||window.opera);
   			if (check) {
  				window.location = arguments[0];
  			} else {
				window.open(arguments[0],'_blank');  				
  			}"$			
			ws.Eval(s, Array As Object(TargetUrl))
		Else
			ws.Eval("window.location = arguments[0]", Array As Object(TargetUrl))
		End If
		ws.Flush
	End If
End Sub

' Parses the event coming from the BROWSER arriving in BANano_ParseEvent. It will redirect the calls to the proper B4J methods.
' 
' self - the class that has the method, typically Me
' ws - the SERVER websocket
' BANPageID - the current pageID.  If the event is beforeunload (internal), then it will be removed from CachedPages
' BANSessionID - the current session ID
' Params - the Params parameter you received in BANano_ParseEvent
'
' Typical usage:
' public Sub BANano_ParseEvent(params As Map)
'	  Main.Server.ParseEvent(Me, ws, CacheReport.BANPageID, params)
' End Sub
public Sub ParseEvent(self As Object, ws As WebSocket, BANPageID As String, BANSessionID As String, Params As Map)
	Dim eventName As String = Params.Get("eventname")
	Dim eventParams() As String = Regex.Split(",",Params.Get("eventparams"))
	Dim RequestID As Int = Params.GetDefault("request_id", -1)	
	'Log(eventName)
	If eventName = "beforeunload" Then
		BANano.B4JRemoveMeFromCache(mCachedPages, BANPageID)		
		Return
	End If
	
	If SubExists(self, eventName) Then
		Params.Remove("eventname")
		Params.Remove("eventparams")
		Params.Remove("request_id")
		If RequestID <> -1 Then
			Dim returnValue As Object
			Select Case Params.Size
				Case 0
					returnValue = CallSub(self, eventName)
				Case 1
					returnValue = CallSub2(self, eventName, Params.Get(eventParams(0)))
				Case 2
					returnValue = CallSub3(self, eventName, Params.Get(eventParams(0)), Params.Get(eventParams(1)))
				Case Else
					Dim ParamsList(Params.Size) As Object
					For i = 0 To Params.Size - 1
						ParamsList(i) = Params.Get(eventParams(i))
					Next
					returnValue = BANano.B4JCallSubX(self, eventName, ParamsList)
			End Select
			
			If returnValue Is String Then
				Dim tmpReturnValue As String = returnValue
				If tmpReturnValue.StartsWith("{") Then ' it is a json string
					ws.eval($"{"request_id": ${RequestID}, "data": ${tmpReturnValue}}"$, Null)
				Else
					ws.eval($"{"request_id": ${RequestID}, "data": "${returnValue}"}"$, Null)
				End If
				'Log(returnValue)
				'Log(" -> is a string")				
				'ws.eval($"{"request_id": ${RequestID}, "data": "${returnValue}"}"$, Null)				
			Else
				'Log(returnValue)
				'Log(" -> is NOT a string")
				ws.eval($"{"request_id": ${RequestID}, "data": ${returnValue}}"$, Null)
			End If			
			'ws.eval($"{"request_id": ${RequestID}, "data": ${returnValue}}"$, Null)
			ws.Flush
		Else
			Select Case Params.Size
				Case 0
					CallSub(self, eventName)
				Case 1
					CallSub2(self, eventName, Params.Get(eventParams(0)))
				Case 2
					CallSub3(self, eventName, Params.Get(eventParams(0)), Params.Get(eventParams(1)))
				Case Else
					Dim ParamsList(Params.Size) As Object
					For i = 0 To Params.Size - 1
						ParamsList(i) = Params.Get(eventParams(i))
					Next
					BANano.B4JCallSubX(self, eventName, ParamsList)
			End Select
		End If	
	End If
End Sub

' Method to cache and retrieve a B4J Server class.
' 
' returns a BANanoCacheReport
'
' <code>
' Sub Class_Globals
' 	Private ws As WebSocket
' 	Private CacheReport As BANanoCacheReport
' End Sub
' 
' Private Sub WebSocket_Connected (WebSocket1 As WebSocket)
' 	ws = WebSocket1
' 	
' 	CacheReport = Main.Server.UpdateFromCache(Me, ws)
' 	Log(CacheReport.BANPageID)
' 	Log(CacheReport.ComesFromCache)
' 	Log(CacheReport.IsReconnected)
' End Sub
' </code>
public Sub UpdateFromCache(B4JClass As Object, ws As WebSocket) As BANanoCacheReport
	Return BANano.B4JUpdateFromCache(B4JClass, mCachedPages, ws)
End Sub

' Use when the BROWSERS websocket is initialized with Initialize
'
' Send a message to the browser telling the server side is ready for communication.  
' It will arrive in the ws_OnServerReady event in the browser class.
' Use if you want to use the OnOpen, OnMessage, OnServerReady, ... events
Public Sub SendReady(ws As WebSocket, eventNameWebsocketInBrowserClass As String)
	ws.RunFunction(eventNameWebsocketInBrowserClass & "_onServerReady", Null)
	ws.Flush
End Sub

'use when the BROWSERS websocket is initialized with InitializeClassic
'
' Send a message to the browser telling the server side is ready for communication.  
' It will arrive in the WebSocket_Connected event in the browser class.
' Use if you use the Websocket_Connected/Disconnected events
public Sub SendConnected(ws As WebSocket)
	ws.RunFunction("websocket_connected", Null)
	ws.Flush
End Sub
